%{ open Ast %}

%token COMMA DBS LPRN RPRN LABS RABS LBRC RBRC LFLOOR RFLOOR LCEIL RCEIL AMP LCASE RCASE PCT LSPLIT RSPLIT
%token SUP SUB
%token ADD MINUS MUL DIV
%token LEQ GEQ NEQ MID NMID EQ LT GT
%token NEG WEDGE VEE
%token SUM FRAC PROD BINOM
%token MAX MIN GCD
%token SQRT
%token SIN COS TAN COT CSC SEC SINH TANH COTH COSH ARCSIN ARCCOS ARCTAN
%token MOD
%token LG LN LOG
%token <int> LIT_INT DIGIT 
%token <float> LIT_FL
%token <string> ID FID
%token EOF EOL

%start program
%type <Ast.program> program

%%

program:
  stmt_list EOF { List.rev $1 }

ident:
| ID { SId($1) }
| ID SUB DIGIT { SSId($1, $3) }

stmt:
  expr_calc DBS { Expr($1) }
| ident EQ expr_calc DBS { Assign($1, $3) }
| PCT expr_calc EOL { Print($2) }
| ident LPRN RPRN EQ stmt { FuncDef($1, [], $5)}
| fun_def RPRN EQ stmt { let (id, args) = $1 in FuncDef(id, List.rev args, $4) }
| expr_spec EQ stmt { let (id, arg) = $1 in FuncDef(id, [arg], $3) }
| fun_def RPRN DBS         { let (id, args) = $1 in Expr(ExprNoId(Call(FId id, List.map (fun a->Id a) (List.rev args)))) }
| ident EQ fun_def RPRN DBS { let (id, args) = $3 in Assign($1, ExprNoId(Call(FId id, List.map (fun a->Id a) (List.rev args)))) }
| PCT fun_def RPRN EOL { let (id, args) = $2 in Print(ExprNoId(Call(FId id, List.map (fun a->Id a) (List.rev args)))) }
| LCASE sutie_list RCASE { Case(List.rev $2) }
| LSPLIT stmt_list RSPLIT { StmtClosure(List.rev $2) }

stmt_list:
    stmt { [$1] }
  | stmt_list stmt { $2 :: $1 }

short_atom:
  DIGIT { ShortLiteral($1) }

long_atom:
  LIT_INT { IntLiteral($1) }
| LIT_FL { FloatLiteral($1) }
| LABS expr_calc RABS { ParenthLike(Abs, $2) }
| LFLOOR expr_calc RFLOOR { ParenthLike(Floor, $2) }
| LCEIL expr_calc RCEIL { ParenthLike(Ceil, $2) }

paren_no_id:
| LPRN expr_no_id RPRN { ExprNoId($2) }
| LPRN expr_spec RPRN { SpecImplId($2) }

paren_id:
| LPRN ident RPRN { $2 }

atom:
  short_atom { ShortAtom($1) }
| long_atom { LongAtom($1) }

atom_closure:
 LBRC expr_calc RBRC { $2 }

expr_pow:
  atom { Atom($1) }
| atom SUP atom_closure { PowerFuncAtom($1, $3) }
| ident SUP atom_closure { PowerFuncId(Id($1), $3) }

expr_log:
  expr_pow { ExprPow($1) }
| LOG SUB atom_closure expr_log { Log($3, $4) }
| LG expr_log { LogLike(Lg, $2) }
| LN expr_log { LogLike(Ln, $2) }
| SQRT expr_log { LogLike(Sqrt, $2) }
| SIN expr_log { LogLike(Sin, $2) }
| COS expr_log { LogLike(Cos, $2) }
| TAN expr_log { LogLike(Tan, $2) }
| ARCSIN expr_log { LogLike(Arcsin, $2) }
| ARCCOS expr_log { LogLike(Arccos, $2) }
| ARCTAN  expr_log { LogLike(Arctan, $2) }
| SINH  expr_log { LogLike(Sinh, $2) }
| TANH  expr_log { LogLike(Tanh, $2) }
| COSH  expr_log { LogLike(Cosh, $2) }
| CSC  expr_log { LogLike(Csc, $2) }
| SEC  expr_log { LogLike(Sec, $2) }
| COT  expr_log { LogLike(Cot, $2) }
| COTH  expr_log { LogLike(Coth, $2) }
| LOG SUB atom_closure ident { LogC($3, Id($4)) }
| LG ident { LogLikeC(Lg, Id($2)) }
| LN ident { LogLikeC(Ln, Id($2)) }
| SQRT ident { LogLikeC(Sqrt, Id($2)) }
| SIN ident { LogLikeC(Sin, Id($2)) }
| COS ident { LogLikeC(Cos, Id($2)) }
| TAN ident { LogLikeC(Tan, Id($2)) }
| ARCSIN ident { LogLikeC(Arcsin, Id($2)) }
| ARCCOS ident { LogLikeC(Arccos, Id($2)) }
| ARCTAN  ident { LogLikeC(Arctan, Id($2)) }
| SINH  ident { LogLikeC(Sinh, Id($2)) }
| TANH  ident { LogLikeC(Tanh, Id($2)) }
| COSH  ident { LogLikeC(Cosh, Id($2)) }
| CSC  ident { LogLikeC(Csc, Id($2)) }
| SEC  ident { LogLikeC(Sec, Id($2)) }
| COT  ident { LogLikeC(Cot, Id($2)) }
| COTH  ident { LogLikeC(Coth, Id($2)) }
| LOG SUB atom_closure paren_no_id { LogC($3, $4) }
| LG paren_no_id { LogLikeC(Lg, $2 ) }
| LN paren_no_id { LogLikeC(Ln, $2 ) }
| SQRT paren_no_id { LogLikeC(Sqrt, $2 ) }
| SIN paren_no_id { LogLikeC(Sin, $2 ) }
| COS paren_no_id { LogLikeC(Cos, $2 ) }
| TAN paren_no_id { LogLikeC(Tan, $2 ) }
| ARCSIN paren_no_id { LogLikeC(Arcsin, $2 ) }
| ARCCOS paren_no_id { LogLikeC(Arccos, $2 ) }
| ARCTAN  paren_no_id { LogLikeC(Arctan, $2 ) }
| SINH  paren_no_id { LogLikeC(Sinh, $2 ) }
| TANH  paren_no_id { LogLikeC(Tanh, $2 ) }
| COSH  paren_no_id { LogLikeC(Cosh, $2 ) }
| CSC  paren_no_id { LogLikeC(Csc, $2 ) }
| SEC  paren_no_id { LogLikeC(Sec, $2 ) }
| COT  paren_no_id { LogLikeC(Cot, $2 ) }
| COTH  paren_no_id { LogLikeC(Coth, $2 ) }
| LOG SUB atom_closure paren_id { LogC($3, Id($4)) }
| LG paren_id { LogLikeC(Lg, Id($2) ) }
| LN paren_id { LogLikeC(Ln, Id($2) ) }
| SQRT paren_id { LogLikeC(Sqrt, Id($2) ) }
| SIN paren_id { LogLikeC(Sin, Id($2) ) }
| COS paren_id { LogLikeC(Cos, Id($2) ) }
| TAN paren_id { LogLikeC(Tan, Id($2) ) }
| ARCSIN paren_id { LogLikeC(Arcsin, Id($2) ) }
| ARCCOS paren_id { LogLikeC(Arccos, Id($2) ) }
| ARCTAN  paren_id { LogLikeC(Arctan, Id($2) ) }
| SINH  paren_id { LogLikeC(Sinh, Id($2) ) }
| TANH  paren_id { LogLikeC(Tanh, Id($2) ) }
| COSH  paren_id { LogLikeC(Cosh, Id($2) ) }
| CSC  paren_id { LogLikeC(Csc, Id($2) ) }
| SEC  paren_id { LogLikeC(Sec, Id($2) ) }
| COT  paren_id { LogLikeC(Cot, Id($2) ) }
| COTH  paren_id { LogLikeC(Coth, Id($2) ) }

expr_impl_mult:
  expr_log { ExprLog($1) }
| expr_impl_mult expr_log { ImplMult($1, $2) }
| expr_impl_mult ident { ImplMultMC($1, Id($2)) }
| expr_impl_mult paren_no_id { ImplMultMC($1, $2) }
| expr_impl_mult paren_id { ImplMultMC($1, Id($2)) }
| ident expr_log { ImplMultC(Id($1), $2) }
| ident ident {ImplMultCMC(Id($1), Id($2))}
| ident paren_no_id { SpecImplNoId($1, $2) }
| paren_no_id expr_log { ImplMultC($1, $2) }
| paren_no_id ident { ImplMultCMC($1, Id($2)) }
| paren_no_id paren_no_id { ImplMultCMC($1, $2) }
| paren_no_id paren_id { ImplMultCMC($1, Id($2)) }
| paren_id expr_log { ImplMultC(Id($1), $2) }
| paren_id ident { ImplMultCMC(Id($1), Id($2)) }
| paren_id paren_no_id { ImplMultCMC(Id($1), $2) }
| paren_id paren_id { ImplMultCMC(Id($1), Id($2)) }

expr_spec:
| ident paren_id { ($1, $2) }

expr_unary:
  expr_impl_mult { ExprImplMult($1) }
| paren_id { Id($1) }
| MINUS expr_unary { Unop(Uminus, $2) }
| ADD expr_unary{ Unop(Uplus, $2) }
| MINUS expr_spec { UnopC(Uminus, SpecImplId($2)) }
| ADD expr_spec { UnopC(Uplus, SpecImplId($2)) }
| MINUS ident { UnopC(Uminus, Id($2)) }
| ADD ident { UnopC(Uplus, Id($2)) }

expr_mult:
  expr_unary { ExprUnary($1) }
| expr_mult MUL expr_unary { MultLike($1, Mult, $3) }
| expr_mult DIV expr_unary { MultLike($1, Div, $3) }
| expr_mult MOD expr_unary { MultLike($1, Mod, $3) }
| expr_mult MUL ident { MultLikeMC($1, Mult, Id($3)) }
| expr_mult DIV ident { MultLikeMC($1, Div, Id($3)) }
| expr_mult MOD ident { MultLikeMC($1, Mod, Id($3)) }
| expr_mult MUL expr_spec { MultLikeMC($1, Mult, SpecImplId($3)) }
| expr_mult DIV expr_spec { MultLikeMC($1, Div, SpecImplId($3)) }
| expr_mult MOD expr_spec { MultLikeMC($1, Mod, SpecImplId($3)) }
| ident MUL expr_unary { MultLikeC(Id($1), Mult, $3) }
| ident DIV expr_unary { MultLikeC(Id($1), Div, $3) }
| ident MOD expr_unary { MultLikeC(Id($1), Mod, $3) }
| ident MUL ident { MultLikeCMC(Id($1), Mult, Id($3)) }
| ident DIV ident { MultLikeCMC(Id($1), Div, Id($3)) }
| ident MOD ident { MultLikeCMC(Id($1), Mod, Id($3)) }
| ident MUL expr_spec { MultLikeCMC(Id($1), Mult, SpecImplId($3)) }
| ident DIV expr_spec { MultLikeCMC(Id($1), Div, SpecImplId($3)) }
| ident MOD expr_spec { MultLikeCMC(Id($1), Mod, SpecImplId($3)) }
| expr_spec MUL expr_unary { MultLikeC(SpecImplId($1), Mult, $3) }
| expr_spec DIV expr_unary { MultLikeC(SpecImplId($1), Div, $3) }
| expr_spec MOD expr_unary { MultLikeC(SpecImplId($1), Mod, $3) }
| expr_spec MUL expr_spec { MultLikeCMC(SpecImplId($1), Mult, SpecImplId($3)) }
| expr_spec DIV expr_spec { MultLikeCMC(SpecImplId($1), Div, SpecImplId($3)) }
| expr_spec MOD expr_spec { MultLikeCMC(SpecImplId($1), Mod, SpecImplId($3)) }
| expr_spec MUL ident { MultLikeCMC(SpecImplId($1), Mult, Id($3)) }
| expr_spec DIV ident { MultLikeCMC(SpecImplId($1), Div, Id($3)) }
| expr_spec MOD ident { MultLikeCMC(SpecImplId($1), Mod, Id($3)) }

expr_add:
  expr_mult { ExprMult($1) }
| expr_add ADD expr_mult {AddLike($1, Plus, $3) }
| expr_add MINUS expr_mult {AddLike($1, Minus, $3) }
| expr_add ADD ident {AddLikeMC($1, Plus, Id($3)) }
| expr_add MINUS ident {AddLikeMC($1, Minus, Id($3)) }
| expr_add ADD expr_spec {AddLikeMC($1, Plus, SpecImplId($3)) }
| expr_add MINUS expr_spec {AddLikeMC($1, Minus, SpecImplId($3)) }
| ident ADD expr_mult {AddLikeC(Id($1), Plus, $3) }
| ident MINUS expr_mult {AddLikeC(Id($1), Minus, $3) }
| ident ADD ident {AddLikeCMC(Id($1), Plus, Id($3)) }
| ident MINUS ident {AddLikeCMC(Id($1), Minus, Id($3)) }
| ident ADD expr_spec {AddLikeCMC(Id($1), Plus, SpecImplId($3)) }
| ident MINUS expr_spec {AddLikeCMC(Id($1), Minus, SpecImplId($3)) }
| expr_spec ADD expr_mult {AddLikeC(SpecImplId($1), Plus, $3) }
| expr_spec MINUS expr_mult {AddLikeC(SpecImplId($1), Minus, $3) }
| expr_spec ADD expr_spec {AddLikeCMC(SpecImplId($1), Plus, SpecImplId($3)) }
| expr_spec MINUS expr_spec {AddLikeCMC(SpecImplId($1), Minus, SpecImplId($3)) }
| expr_spec ADD ident {AddLikeCMC(SpecImplId($1), Plus, Id($3)) }
| expr_spec MINUS ident {AddLikeCMC(SpecImplId($1), Minus, Id($3)) }

expr_no_id:
  expr_add { ExprAdd($1) }
| ident LPRN RPRN {Call(FId($1), [])}
| call_expr RPRN { let (id, args) = $1 in Call(id, List.rev args) }
| GCD LPRN arg_list RPRN { Call(Gcd, $3) }
| MAX LPRN arg_list RPRN { Call(Max, $3) }
| MIN LPRN arg_list RPRN { Call(Min, $3) }
| FRAC atom_closure atom_closure { FracLike(Frac, $2, $3) }
| BINOM atom_closure atom_closure { FracLike(Binom, $2, $3) }

fun_def:
| ident LPRN ident COMMA ident       { ($1, [$5; $3]) }
| fun_def COMMA ident     { let (id, args) = $1 in (id, $3 :: args) }

call_expr:
 | ident LPRN ident COMMA expr_no_id { (FId($1), [ExprNoId($5); Id($3)])}
 | ident LPRN expr_no_id COMMA expr_no_id { (FId($1), [ExprNoId($5); ExprNoId($3)])}
 | ident LPRN expr_no_id COMMA ident { (FId($1), [Id($5); ExprNoId($3)])}
 | ident LPRN expr_spec COMMA expr_no_id { (FId($1), [ExprNoId($5); SpecImplId($3)])}
 | ident LPRN expr_spec COMMA expr_spec { (FId($1), [SpecImplId($5); SpecImplId($3)])}
 | ident LPRN expr_no_id COMMA expr_spec { (FId($1), [SpecImplId($5); ExprNoId($3)])}
 | ident LPRN expr_spec COMMA ident { (FId($1), [Id($5); SpecImplId($3)])}
 | ident LPRN ident COMMA expr_spec { (FId($1), [SpecImplId($5); Id($3)])}
 | fun_def COMMA expr_no_id { let (id, args) = $1 in (FId id, ExprNoId($3) :: List.map (fun a->Id a) args) }
 | call_expr COMMA expr_calc { let (id, args) = $1 in (id, $3 :: args) }

expr_calc:
  ident         { Id($1) }
| expr_spec { SpecImplId($1) }
| expr_no_id { ExprNoId($1) }

arg_list:
  expr_calc { [$1] }
| arg_list COMMA expr_calc { $3 :: $1 }

expr_comp:
  expr_calc LT expr_calc { Comp($1, Lt, $3) }
| expr_calc GT expr_calc { Comp($1, Gt, $3) }
| expr_calc LEQ expr_calc { Comp($1, Leq, $3) }
| expr_calc GEQ expr_calc { Comp($1, Geq, $3) }
| expr_calc EQ expr_calc { Comp($1, Eq, $3) }
| expr_calc NEQ expr_calc { Comp($1, Neq, $3) }
| expr_calc NMID expr_calc { Comp($1, Nmid, $3) }
| expr_calc MID expr_calc { Comp($1, Mid, $3) }

expr_logic_atom:
  expr_comp { ExprComp($1) }
| LPRN expr_logic RPRN { ParenLogic($2) }
| NEG expr_logic_atom { Not($2)}

expr_logic:
  expr_logic_atom { LogicAtom($1) }
| expr_logic WEDGE expr_logic_atom { Logic($1, And, $3) }
| expr_logic VEE expr_logic_atom { Logic($1, Or, $3) }

sutie:
  stmt AMP expr_logic DBS { CaseSutie($1, $3) }

sutie_list:
    sutie { [$1] }
  | sutie_list sutie { $2 :: $1 }
  