open OUnit2
open Scanner
open Parser

let token_list str =
 let lexbuf = Lexing.from_string str in
 let rec next l =
   match ctex_read_token lexbuf with
       EOF -> List.rev (EOF :: l)
     | s -> next (s :: l)
 in next [];;

let assert_and_print_on_failure l1 l2 =
  if List.equal (fun a b -> a = b) l1 l2 then assert_equal 0 0 else
    let () = Printf.printf("\n") in
    let () = List.iter (fun t -> print_token t true) l1 in
    let () = Printf.printf("\n") in
    let () = List.iter (fun t -> print_token t true) l2 in
    assert_equal 0 1

let scanner_test_list = [
  ("Empty", "", [ EOF ]) ;
  ("Digit", "1", [ DIGIT 1 ; EOF ]) ;
  ("Backslash should be replaced in ID", "\\mathrm{id}", [ ID "_mathrm___id___" ; EOF ]) ;
  ("Backslash should be replaced in ID", "\\alpha", [ ID "__alpha" ; EOF ]) ;
  ("Complex ID", "\\mathrm{id\\alpha\\beta}", [ ID "_mathrm___id__alpha__beta___" ; EOF ]) ;
  ("ID with underscore", "a_3", [ ID "a" ; SUB ; DIGIT 3 ; EOF ]) ;
]

let scanner_tests = "Test suite for Scanner" >::: List.map (fun (desc, input, result) ->
  desc >:: (fun _ -> assert_and_print_on_failure (token_list input) result)
) scanner_test_list

let _ = run_test_tt_main scanner_tests
